import re
import io

from typing import Callable
from html import unescape, escape


# LLM 的 XML 难以严格符合转义标准，处理它以让它能通过标准的 XML 来格式化
def normal_llm_response_xml(text: str):
  return _process_xml_text(text, lambda x: escape(unescape(x)))

# 一个简单标准的 XML TAG 表达式，以避免和各种复杂的转义规则冲突。
# LLM 只会生成这个简单标准的表达式，因此不符合的部分一定是正文，从而排除它们可能触碰转义规则的可能。
# 1. 所有的 TAG、属性名必须是字母、数字、下划线、中脊线
# 2. 所有的属性值必须是字母、数字、下划线、中脊线、英语逗号
# 3. 属性必须是双引号
_XML_TAG_PATTERN = re.compile(
  pattern=r'<([a-zA-Z0-9_\-]+)(?:\s+[a-zA-Z][a-zA-Z0-9_\-]*="[a-zA-Z0-9_,\-]*")*\s*/?>|</([a-zA-Z0-9_\-]+)>',
)

def _process_xml_text(text: str, process_func: Callable[[str], str]) -> str:
  buffer = io.StringIO()
  prev_end: int = 0

  for match in _XML_TAG_PATTERN.finditer(text):
    non_tag_part = text[prev_end:match.start()]
    if non_tag_part:
      buffer.write(process_func(non_tag_part))

    buffer.write(match.group())
    prev_end = match.end()

  return buffer.getvalue()


from json import dumps
from xml.etree.ElementTree import Element
from ..llm import LLM
from .index import Index
from .utils import normalize_xml_text, parse_page_indexes


def analyse_position(llm: LLM, index: Index | None, chunk_xml: Element) -> Element:
  if index is None:
    return chunk_xml # TODO: implements citations position

  content_xml = chunk_xml.find("content")
  raw_pages_root = Element("pages")
  origin_headlines: list[Element] = []

  for child in content_xml:
    if child.tag != "headline":
      continue

    page_indexes = parse_page_indexes(child)
    if not index.after_first_index_page(page_indexes[0]):
      continue

    headline = Element("headline")
    headline.text = normalize_xml_text(child.text)
    raw_pages_root.append(headline)
    origin_headlines.append(child)

  response_xml = llm.request_xml("position", raw_pages_root, {
    "index": dumps(
      obj=index.json,
      ensure_ascii=False,
      indent=2,
    ),
  })

  for i, headline in enumerate(response_xml):
    id = headline.get("id")
    if id is not None and i < len(origin_headlines):
      origin_headline = origin_headlines[i]
      origin_headline.set("id", id)
      origin_headline.text = normalize_xml_text(headline.text)

  return chunk_xml


from typing import Generator, Callable, Iterable
from dataclasses import dataclass
from xml.etree.ElementTree import tostring, Element
from resource_segmentation import Group, Segment, Resource
from ..llm import LLM
from .common import PageRef


@dataclass
class PageXML:
  page_index: int
  xml: Element
  is_gap: bool = False

def get_and_clip_pages(llm: LLM, group: Group[PageRef], get_element: Callable[[int], Element]) -> list[PageXML]:
  head = _get_pages(
    llm=llm,
    items=group.head,
    remain_tokens=group.head_remain_count,
    clip_tail=False,
    get_element=get_element,
  )
  tail = _get_pages(
    llm=llm,
    items=group.tail,
    remain_tokens=group.tail_remain_count,
    clip_tail=True,
    get_element=get_element,
  )
  body = _get_pages(
    llm=llm,
    items=group.body,
    remain_tokens=None,
    clip_tail=True,
    get_element=get_element,
  )
  page_xml_list: list[PageXML] = []

  for page_xml in reversed(list(head)):
    page_xml.is_gap = True
    page_xml_list.append(page_xml)

  for page_xml in body:
    page_xml_list.append(page_xml)

  for page_xml in tail:
    page_xml.is_gap = True
    page_xml_list.append(page_xml)

  return page_xml_list

def _get_pages(
    llm: LLM,
    items: list[Resource[PageRef] | Segment[PageRef]],
    remain_tokens: int | None,
    clip_tail: bool,
    get_element: Callable[[int], Element],
  ) -> Generator[PageXML, None, None]:

  if len(items) == 0:
    return

  if remain_tokens is not None:
    tokens = sum(item.count for item in items)
    if remain_tokens == tokens:
      remain_tokens = None

  if remain_tokens is None:
    for text_info in _search_resources(items):
      page_index = text_info.payload.page_index
      yield PageXML(
        page_index=page_index,
        xml=get_element(page_index),
      )
  else:
    clipped: list[Resource[PageRef]] = []
    page_xml_list: list[PageXML] = []
    iter_source: Iterable[Resource[PageRef]] = _search_resources(items)
    if not clip_tail:
      iter_source = reversed(list(iter_source))

    for text_info in iter_source:
      if remain_tokens > 0:
        clipped.append(text_info)
      if remain_tokens >= text_info.count:
        remain_tokens -= text_info.count
      else:
        break
    if not clip_tail:
      clipped.reverse()

    for i, resource in enumerate(clipped):
      page_index = resource.payload.page_index
      page_xml: Element | None = get_element(page_index)

      if remain_tokens > 0 and (
        (clip_tail and i == len(clipped) - 1) or \
        (not clip_tail and i == 0)
      ):
        page_xml = _clip_element(
          llm=llm,
          element=page_xml,
          remain_tokens=remain_tokens,
          clip_tail=clip_tail,
        )
      if page_xml is not None:
        page_xml_list.append(PageXML(
          page_index=page_index,
          xml=page_xml,
        ))

    if not clip_tail:
      page_xml_list.reverse()
    yield from page_xml_list

def _search_resources(items: list[Resource[PageRef] | Segment[PageRef]]):
  for item in items:
    if isinstance(item, Resource):
      yield item
    elif isinstance(item, Segment):
      yield from item.resources

def _clip_element(llm: LLM, element: Element, remain_tokens: int, clip_tail: bool) -> Element | None:
  clipped_element = Element(element.tag, element.attrib)
  children: list[tuple[Element, int]] = []
  remain_tokens -= llm.count_tokens_count(tostring(clipped_element, encoding="unicode"))

  for child in element:
    child_text = tostring(child, encoding="unicode")
    child_tokens = llm.count_tokens_count(child_text)
    children.append((child, child_tokens))
  if not clip_tail:
    children.reverse()

  if len(children) == 0:
    text = element.text
    if text is None:
      return None
    tokens = llm.encode_tokens(text)
    if clip_tail:
      tokens = tokens[:remain_tokens]
    else:
      tokens = tokens[len(tokens) - remain_tokens:]
    if len(tokens) == 0:
      return None
    clipped_element.text = llm.decode_tokens(tokens)

  else:
    clipped_children: list[Element] = []
    for child, tokens_count in children:
      if remain_tokens >= tokens_count:
        remain_tokens -= tokens_count
        clipped_children.append(child)
      else:
        child = _clip_element(llm, child, remain_tokens, clip_tail)
        if child is not None:
          clipped_children.append(child)
        break
    if len(clipped_children) == 0:
      return None
    if not clip_tail:
      clipped_children.reverse()
    clipped_element.extend(clipped_children)

  return clipped_element


import re
import json

from typing import cast, Any
from importlib.resources import files
from jinja2 import Environment, Template
from xml.etree.ElementTree import tostring, fromstring, Element
from pydantic import SecretStr
from tiktoken import get_encoding, Encoding
from langchain_core.messages import SystemMessage, HumanMessage

from ..template import create_env
from .increasable import Increasable
from .executor import LLMExecutor
from .escape import normal_llm_response_xml


class LLM:
  def __init__(
      self,
      key: str,
      url: str,
      model: str,
      token_encoding: str,
      timeout: float | None = None,
      top_p: float | tuple[float, float] | None = None,
      temperature: float | tuple[float, float] | None = None,
      retry_times: int = 5,
      retry_interval_seconds: float = 6.0,
    ):
    prompts_path = files("pdf_craft").joinpath("data/prompts")
    self._templates: dict[str, Template] = {}
    self._encoding: Encoding = get_encoding(token_encoding)
    self._env: Environment = create_env(prompts_path)
    self._executor = LLMExecutor(
      url=url,
      model=model,
      api_key=cast(SecretStr, key),
      timeout=timeout,
      top_p=Increasable(top_p),
      temperature=Increasable(temperature),
      retry_times=retry_times,
      retry_interval_seconds=retry_interval_seconds,
    )

  def request_json(self, template_name: str, user_data: Element, params: dict[str, Any] | None = None) -> Any:
    if params is None:
      params = {}
    return self._executor.request(
      input=self._create_input(template_name, user_data, params),
      parser=self._encode_json,
    )

  def request_xml(self, template_name: str, user_data: Element, params: dict[str, Any] | None = None) -> Element:
    if params is None:
      params = {}
    return self._executor.request(
      input=self._create_input(template_name, user_data, params),
      parser=self._encode_xml,
    )

  def _create_input(self, template_name: str, user_data: Element, params: dict[str, Any]):
    template = self._template(template_name)
    data = tostring(user_data, encoding="unicode")
    prompt = template.render(**params)
    return [
      SystemMessage(content=prompt),
      HumanMessage(content=data)
    ]

  def prompt_tokens_count(self, template_name: str, params: dict[str, Any]) -> int:
    template = self._template(template_name)
    prompt = template.render(**params)
    return len(self._encoding.encode(prompt))

  def encode_tokens(self, text: str) -> list[int]:
    return self._encoding.encode(text)

  def decode_tokens(self, tokens: list[int]) -> str:
    return self._encoding.decode(tokens)

  def count_tokens_count(self, text: str) -> int:
    return len(self._encoding.encode(text))

  def _template(self, template_name: str) -> Template:
    template = self._templates.get(template_name, None)
    if template is None:
      template = self._env.get_template(template_name)
      self._templates[template_name] = template
    return template

  def _encode_json(self, response: str) -> Any:
    response = re.sub(r"^```JSON", "", response)
    response = re.sub(r"```$", "", response)
    try:
      return json.loads(response)
    except Exception as e:
      print(response)
      raise e

  def _encode_xml(self, response: str) -> Element:
    response = re.sub(r"^```XML", "", response)
    response = re.sub(r"```$", "", response)
    response = normal_llm_response_xml(response)
    try:
      return fromstring(response)
    except Exception as e:
      print(response)
      raise e


from dataclasses import dataclass


@dataclass
class LLMWindowTokens:
  main_texts: int | None = None
  citations: int | None = None

def parse_window_tokens(window_tokens: LLMWindowTokens | int | None) -> LLMWindowTokens:
  main_texts = 4400
  citations = 3300

  if isinstance(window_tokens, int):
    main_texts = window_tokens
    citations = window_tokens

  elif isinstance(window_tokens, LLMWindowTokens):
    if window_tokens.main_texts is not None:
      main_texts = window_tokens.main_texts
    if window_tokens.citations is not None:
      citations = window_tokens.citations

  return LLMWindowTokens(
    main_texts=main_texts,
    citations=citations,
  )


import os
import fitz

from html import escape
from typing import Generator, Iterable
from PIL.Image import Image
from xml.etree.ElementTree import fromstring, Element, ParseError

from .types import AnalysingStep, AnalysingProgressReport, AnalysingStepReport
from .asset_matcher import search_asset_tags, AssetMatcher, AssetKind
from ..utils import sha256_hash
from ..pdf import (
  PDFPageExtractor,
  Block,
  Text,
  TextKind,
  TextBlock,
  TableBlock,
  TableFormat,
  FigureBlock,
  FormulaBlock,
)


def extract_ocr_page_xmls(
    extractor: PDFPageExtractor,
    pdf_path: str,
    expected_page_indexes: set[int],
    cover_path: str,
    assets_dir_path: str,
    report_step: AnalysingStepReport | None,
    report_progress: AnalysingProgressReport | None,
  ) -> Generator[Element, None, None]:

  with fitz.open(pdf_path) as pdf:
    if report_step is not None:
      report_step(
        AnalysingStep.OCR,
        pdf.page_count - len(expected_page_indexes),
      )
    for i, blocks, image in extractor.extract_enumerated_blocks_and_image(
      pdf=pdf,
      page_indexes=(i for i in range(pdf.page_count) if i not in expected_page_indexes),
    ):
      if i == 0:
        image.save(cover_path)

      page_xml = _transform_page_xml(blocks)
      _migrate_expressions_and_save_images(
        root=page_xml,
        blocks=blocks,
        assets_dir_path=assets_dir_path,
      )
      yield i, page_xml

      if report_progress is not None:
        report_progress(i + 1)

def _transform_page_xml(blocks: list[Block]) -> Element:
  root = Element("page")
  for block in blocks:
    if isinstance(block, TextBlock):
      tag_name: str
      if block.kind == TextKind.TITLE:
        tag_name = "headline"
      elif block.kind == TextKind.PLAIN_TEXT:
        tag_name = "text"
      elif block.kind == TextKind.ABANDON:
        tag_name = "abandon"

      text_dom = Element(tag_name)
      if block.kind == TextKind.PLAIN_TEXT:
        text_dom.set("indent", "true" if block.has_paragraph_indentation else "false")
        text_dom.set("touch-end", "true" if block.last_line_touch_end else "false")

      _extends_line_doms(text_dom, block.texts)
      root.append(text_dom)

    elif isinstance(block, TableBlock):
      _append_asset_dom(root, block, "table")

    elif isinstance(block, FormulaBlock):
      _append_asset_dom(root, block, "formula")

    elif isinstance(block, FigureBlock):
      _append_asset_dom(root, block, "figure")

  return root

def _append_asset_dom(root: Element, block: Block, tag_name: str):
  root.append(Element(tag_name))
  if len(block.texts) > 0:
    caption_dom = Element(f"{tag_name}-caption")
    _extends_line_doms(caption_dom, block.texts)
    root.append(caption_dom)

def _extends_line_doms(parent: Element, texts: list[Text]):
  for text in texts:
    content = text.content.replace("\n", " ")
    content = escape(content.strip())
    line_dom = Element("line")
    line_dom.set("confidence", "{:.2f}".format(text.rank))
    line_dom.text = content
    parent.append(line_dom)

def _migrate_expressions_and_save_images(root: Element, blocks: list[Block], assets_dir_path: str):
  asset_matcher = AssetMatcher()
  images: dict[str, Image] = {}

  def register_image_and_get_hash(image: Image):
    hash = sha256_hash(image.tobytes())
    images[hash] = image
    return hash

  def create_children(tag_name: str, text: str) -> Iterable[Element]:
    child = Element(tag_name)
    child.text = text
    return (child,)

  for block in blocks:
    kind: AssetKind | None = None
    hash: str | None = None
    children: Iterable[Element] | None = None

    if isinstance(block, TableBlock):
      kind = AssetKind.TABLE
      hash = register_image_and_get_hash(block.image)
      if block.format == TableFormat.LATEX:
        children = create_children("latex", block.content)
      elif block.format == TableFormat.MARKDOWN:
        children = create_children("markdown", block.content)
      elif block.format == TableFormat.HTML:
        try:
          table_element = fromstring(block.content)
          table_content = Element("html")
          table_content.append(table_element)
          children = (table_content,)
        except ParseError:
          pass

    elif isinstance(block, FormulaBlock):
      kind = AssetKind.FORMULA
      hash = register_image_and_get_hash(block.image)
      if block.content is not None:
        children = create_children("latex", block.content)

    elif isinstance(block, FigureBlock):
      kind = AssetKind.FIGURE
      hash = register_image_and_get_hash(block.image)

    if kind is not None:
      asset_matcher.register_virtual_dom(
        kind=kind,
        hash=hash,
        children=children,
      )

  asset_matcher.recover_asset_doms_for_xml(root)

  for asset_dom in search_asset_tags(root):
    hash = asset_dom.get("hash", None)
    if hash is None:
      continue
    image: Image | None = images.get(hash, None)
    if image is None:
      continue
    file_path = os.path.join(assets_dir_path, f"{hash}.png")
    if os.path.exists(file_path):
      continue
    image.save(file_path, "PNG")



class Increaser:
  def __init__(self, value_range: tuple[float, float] | None):
    self._value_range: tuple[float, float] | None = value_range
    self._current: float | None = value_range[0] if value_range is not None else None

  @property
  def current(self) -> float | None:
    return self._current

  def increase(self):
    if self._value_range is None:
      return
    _, end_value = self._value_range
    self._current = self._current + 0.5 * (end_value - self._current)

class Increasable:
  def __init__(self, param: float | tuple[float, float] | None):
    self._value_range: tuple[float, float] | None = None

    if isinstance(param, int):
      param = float(param)
    if isinstance(param, float):
      param = (param, param)
    if isinstance(param, tuple):
      if len(param) != 2:
        raise ValueError(f"Expected a tuple of length 2, got {len(param)}")
      begin, end = param
      if isinstance(begin, int):
        begin = float(begin)
      if isinstance(end, int):
        end = float(end)
      self._value_range = (begin, end)

  def context(self) -> Increaser:
    return Increaser(self._value_range)


from __future__ import annotations

from typing import Iterable, Sequence, Callable
from doc_page_extractor import Point, Rectangle, Layout, OCRFragment
from .text_matcher import check_texts_matching_rate
from .utils import rate, intersection_area_size

class _Shape:
  def __init__(self, layout: Layout):
    self.layout: Layout = layout
    self.pre: list[Layout | None] = [None, None]
    self.nex: list[Layout | None] = [None, None]

  @property
  def distance2(self) -> float:
    x, y = self.layout.rect.lt
    return x*x + y*y

class Section:
  def __init__(
      self,
      page_index: int,
      layouts: Iterable[Layout],
    ) -> list[Layout]:
    self._page_index: int = page_index
    self._shapes: list[_Shape] = [_Shape(layout) for layout in layouts]

  @property
  def page_index(self) -> int:
    return self._page_index

  def framework(self) -> list[Layout]:
    pre_shapes = list(self._side_framework(lambda shape: shape.pre))
    nex_shapes = list(self._side_framework(lambda shape: shape.nex))
    framework_shapes: list[_Shape]

    if len(pre_shapes) < len(nex_shapes):
      framework_shapes = nex_shapes
    else:
      framework_shapes = pre_shapes

    return [shape.layout for shape in framework_shapes]

  def _side_framework(self, get_side: Callable[[_Shape], list[Layout | None]]):
    for shape in self._shapes:
      matched_shapes = get_side(shape)
      # The logic here should be understood as the processing of book page scans.
      # When the book is opened, the even-numbered pages will be on the left and the odd-numbered pages will be on the right.
      # Usually when editing and typesetting, even-numbered pages will be matched with one typesetting format,
      # and odd-numbered pages will be matched with another typesetting format, in order to achieve aesthetics.
      # Therefore, typesetting format (framework) matching usually takes into account the situation of matching every other page.
      if matched_shapes[0] is not None or matched_shapes[1] is not None:
        yield shape

  def link_next(self, next: Section, offset: int) -> None:
    assert offset in (1, 2), f"invalid offset {offset}"
    matched_shapes_matrix: list[list[_Shape]] = []
    for shape in self._shapes:
      matched_shapes_matrix.append([
        # pylint: disable=W0212
        next_shape for next_shape in next._shapes
        if self._is_shape_contents_matches(shape, next_shape)
      ])

    origin_shapes = self._find_origin_shapes(matched_shapes_matrix)
    if origin_shapes is not None:
      origins = (origin_shapes[0].layout.rect.lt, origin_shapes[1].layout.rect.lt)
      for shape1, shape2 in self._iter_matched_shapes(origins, matched_shapes_matrix):
        shape1.nex[offset - 1] = shape2
        shape2.pre[offset - 1] = shape1

  def _is_shape_contents_matches(self, shape1: _Shape, shape2: _Shape) -> bool:
    size_match_rate = 0.95
    layout1 = shape1.layout
    layout2 = shape2.layout
    size1 = layout1.rect.size
    size2 = layout2.rect.size
    if rate(size1[0], size2[0]) < size_match_rate or \
       rate(size1[1], size2[1]) < size_match_rate:
      return False

    matched_count: int = 0

    for fragment1 in layout1.fragments:
      for fragment2 in layout2.fragments:
        if self._is_fragments_matches(layout1, layout2, fragment1, fragment2):
          matched_count += 1
          break

    fragments_count = max(len(layout1.fragments), len(layout2.fragments))
    if fragments_count == 0:
      return True

    return self._check_group_matches(
       matched_count / fragments_count,
      fragments_count,
      (0.0, 0.45, 0.45, 0.6, 0.8, 0.95),
    )

  def _is_fragments_matches(self, layout1: Layout, layout2: Layout, fragment1: OCRFragment, fragment2: OCRFragment) -> bool:
    size_match_rate = 0.85
    rect1 = self._relative_rect(layout1.rect.lt, fragment1.rect)
    rect2 = self._relative_rect(layout2.rect.lt, fragment2.rect)
    if self._intersection_rate(rect1, rect2) < size_match_rate:
      return False

    text_rate, text_length = check_texts_matching_rate(fragment1.text, fragment2.text)
    return self._check_group_matches(
      text_rate,
      text_length,
      (0.0, 0.0, 0.5, 0.55, 0.6, 0.75, 0.8, 0.95),
    )

  def _find_origin_shapes(self, matched_shapes: list[list[_Shape]]):
    origin_shape1: _Shape | None = None
    origin_shape2: _Shape | None = None
    origin_matched_shape: list[_Shape] | None = None
    min_distance2: float = float("inf")

    for shape, matched_shapes in zip(self._shapes, matched_shapes):
      if len(matched_shapes) == 0:
        continue
      distance2 = shape.distance2
      if distance2 < min_distance2:
        origin_shape1 = shape
        origin_matched_shape = matched_shapes
        min_distance2 = distance2

    if origin_shape1 is None:
      return None

    assert origin_matched_shape is not None
    min_distance2 = float("inf")

    for shape in origin_matched_shape:
      distance2 = shape.distance2
      if distance2 < min_distance2:
        origin_shape2 = shape
        min_distance2 = distance2

    assert origin_shape2 is not None
    return origin_shape1, origin_shape2

  def _iter_matched_shapes(self, origins: tuple[Point, Point], matched_shapes_matrix: list[list[_Shape]]):
    for shape1, matched_shapes in zip(self._shapes, matched_shapes_matrix):
      if len(matched_shapes) == 0:
        continue

      rect1 = self._relative_rect(origins[0], shape1.layout.rect)
      max_area_rate: float = float("-inf")
      matched_shape2: _Shape | None = None

      for shape2 in matched_shapes:
        rect2 = self._relative_rect(origins[1], shape2.layout.rect)
        size_rate = self._intersection_rate(rect1, rect2)
        if size_rate > max_area_rate:
          max_area_rate = size_rate
          matched_shape2 = shape2

      if max_area_rate >= 0.85:
        assert matched_shape2 is not None
        yield shape1, matched_shape2

  def _check_group_matches(self, calculated_rate: float, group_count: int, rates_list: Sequence[float]) -> bool:
    if group_count < len(rates_list):
      return calculated_rate >= rates_list[group_count]
    else:
      return calculated_rate >= rates_list[-1]

  def _relative_rect(self, origin: Point, rect: Rectangle) -> Rectangle:
    return Rectangle(
      lt=(rect.lt[0] - origin[0], rect.lt[1] - origin[1]),
      rt=(rect.rt[0] - origin[0], rect.rt[1] - origin[1]),
      lb=(rect.lb[0] - origin[0], rect.lb[1] - origin[1]),
      rb=(rect.rb[0] - origin[0], rect.rb[1] - origin[1])
    )

  def _intersection_rate(self, rect1: Rectangle, rect2: Rectangle) -> float:
    width1, height1 = rect1.size
    width2, height2 = rect2.size
    width, height = intersection_area_size(rect1, rect2)
    width_rate = width / max(width1, width2)
    height_rate = height / max(height1, height2)
    return 0.5 * (width_rate + height_rate)



from xml.etree.ElementTree import Element
from ..llm import LLM
from .asset_matcher import search_asset_tags, AssetMatcher, ASSET_TAGS


def analyse_page(llm: LLM, raw_page_xml: Element, previous_page_xml: Element | None):
  if len(raw_page_xml) == 0:
    # some page from PDF maybe empty
    return Element("page")

  if previous_page_xml is None:
    raw_page_xml.set("previous-page", "null")
  elif previous_page_xml.tag == "index":
    raw_page_xml.set("previous-page", "index")
  else:
    raw_page_xml.set("previous-page", "page")

  asset_matcher = AssetMatcher().register_raw_xml(raw_page_xml)
  raw_page_xml = _clean_hash_from_assets(raw_page_xml)
  response_xml = llm.request_xml("page", raw_page_xml)

  if response_xml.tag == "index":
    index_xml = Element("index")
    index_xml.extend(raw_page_xml)
    return index_xml

  elif response_xml.tag == "page":
    asset_matcher.recover_asset_doms_for_xml(response_xml)
    _collect_for_citation(response_xml)
    return response_xml

def _clean_hash_from_assets(xml: Element) -> Element:
  for asset_xml in search_asset_tags(xml):
    asset_xml.attrib = {}
  return xml

def _collect_for_citation(response_root: Element):
  citation: Element | None = None
  for child in list(response_root):
    if child.tag == "citation":
      citation = child
    elif citation is not None:
      citation.append(child)
      response_root.remove(child)

def _handle_asset_tags(parent: Element):
  pre_asset: Element | None = None
  asset_captions: list[Element] = []
  for child in parent:
    if child.tag not in ASSET_TAGS:
      if child.tag == "citation":
        _handle_asset_tags(child)
      if pre_asset is not None and \
         child.tag == f"{pre_asset.tag}-caption":
        for caption_child in child:
          pre_asset.append(caption_child)
        asset_captions.append(child)
      pre_asset = None
    if "hash" in child.attrib:
      pre_asset = child
      yield child.get("hash")
  for asset_caption in asset_captions:
    parent.remove(asset_caption)



from typing import Callable
from enum import auto, Enum


class AnalysingStep(Enum):
  OCR = auto()
  ANALYSE_PAGE = auto()
  EXTRACT_INDEX = auto()
  EXTRACT_CITATION = auto()
  EXTRACT_MAIN_TEXT = auto()
  MARK_POSITION = auto()
  ANALYSE_META = auto()
  GENERATE_CHAPTERS = auto()

# func(completed_count: int) -> None
AnalysingProgressReport = Callable[[int], None]

# func(step: AnalysingStep, count: int) -> None
AnalysingStepReport = Callable[[AnalysingStep, int], None]



from __future__ import annotations
import os

from json import loads
from dataclasses import dataclass
from typing import Callable
from xml.etree.ElementTree import tostring, Element
from .i18n import I18N
from .template import Template


@dataclass
class NavPoint:
  index_id: int
  order: int
  file_name: str

def gen_index(
    template: Template,
    i18n: I18N,
    meta: dict,
    index_file_path: str,
    has_cover: bool,
    check_chapter_exits: Callable[[int], bool],
  ) -> tuple[str, list[NavPoint]]:

  nav_elements: list[Element]
  nav_points: list[NavPoint]
  depth: int

  if os.path.exists(index_file_path):
    prefaces, chapters = _parse_index(index_file_path)
    nav_point_generation = _NavPointGeneration(
      has_cover=has_cover,
      check_chapter_exits=check_chapter_exits,
      chapters_count=(
        _count_chapters(prefaces) +
        _count_chapters(chapters)
      ),
    )
    nav_elements = []
    for chapters_list in (prefaces, chapters):
      for chapter in chapters_list:
        element = nav_point_generation.generate(chapter)
        if element is not None:
          nav_elements.append(element)

    depth = max(
      _max_depth(prefaces),
      _max_depth(chapters),
    )
    nav_points = nav_point_generation.nav_points

  else:
    nav_elements = []
    nav_points = []
    depth = 0

  toc_ncx = template.render(
    template="toc.ncx",
    depth=depth,
    i18n=i18n,
    meta=meta,
    has_cover=has_cover,
    nav_points=[tostring(p, encoding="unicode") for p in nav_elements],
  )
  return toc_ncx, nav_points

def _count_chapters(chapters: list[_Chapter]) -> int:
  count: int = 0
  for chapter in chapters:
    count += 1 + _count_chapters(chapter.children)
  return count

def _max_depth(chapters: list[_Chapter]) -> int:
  max_depth: int = 0
  for chapter in chapters:
    max_depth = max(
      max_depth,
      _max_depth(chapter.children) + 1,
    )
  return max_depth

class _NavPointGeneration:
  def __init__(self, has_cover: bool, chapters_count: int, check_chapter_exits: Callable[[int], bool]):
    self._nav_points: list[NavPoint] = []
    self._next_order: int = 2 if has_cover else 1
    self._digits = len(str(chapters_count))
    self._check_chapter_exits: Callable[[int], bool] = check_chapter_exits

  @property
  def nav_points(self) -> list[NavPoint]:
    return self._nav_points

  def generate(self, chapter: _Chapter) -> Element | None:
    if not self._check_chapter_exits(chapter.id):
      return None

    part_id = str(chapter.id).zfill(self._digits)
    file_name = f"part{part_id}.xhtml"
    order = self._next_order

    nav_point_xml = Element("navPoint")
    nav_point_xml.set("id", f"np_{chapter.id}")
    nav_point_xml.set("playOrder", str(order))

    label_xml = Element("navLabel")
    label_text_xml = Element("text")
    label_text_xml.text = chapter.headline
    label_xml.append(label_text_xml)

    content_xml = Element("content")
    content_xml.set("src", f"Text/{file_name}")

    nav_point_xml.append(label_xml)
    nav_point_xml.append(content_xml)

    self._next_order += 1
    self._nav_points.append(NavPoint(
      index_id=chapter.id,
      order=order,
      file_name=file_name,
    ))
    for child in chapter.children:
      child_xml = self.generate(child)
      if child_xml is not None:
        nav_point_xml.append(child_xml)

    return nav_point_xml

@dataclass
class _Chapter:
  id: int
  headline: str
  children: list[_Chapter]

def _parse_index(file_path: str) -> tuple[list[_Chapter], list[_Chapter]]:
  data: dict | list
  with open(file_path, "r", encoding="utf-8") as file:
    data = loads(file.read())
  if isinstance(data, list):
    return [], _transform_chapters(data)
  elif isinstance(data, dict):
    return (
      _transform_chapters(data["prefaces"]),
      _transform_chapters(data["chapters"]),
    )

def _transform_chapters(data_list: list) -> list[_Chapter]:
  chapters: list[_Chapter] = []
  for data in data_list:
    chapters.append(_Chapter(
      id=int(data["id"]),
      headline=data["headline"],
      children=_transform_chapters(data["children"]),
    ))
  return chapters


import os

from zipfile import ZipFile
from .types import TableRender, LaTeXRender


class Context:
  def __init__(
        self,
        file: ZipFile,
        assets_path: str | None,
        table_render: TableRender,
        latex_render: LaTeXRender,
      ) -> None:

    if assets_path is not None and not os.path.exists(assets_path):
      assets_path = None
    self._assets_path: str | None = assets_path
    self._file: ZipFile = file
    self._table_render: TableRender = table_render
    self._latex_render: LaTeXRender = latex_render
    self._used_file_names: dict[str, str] = {}
    self._asset_files: list[str] = []

    if assets_path is not None:
      for file in os.listdir(assets_path):
        if not file.startswith("."):
          self._asset_files.append(file)
      self._asset_files.sort()

  @property
  def file(self) -> ZipFile:
    return self._file

  @property
  def table_render(self) -> TableRender:
    return self._table_render

  @property
  def latex_render(self) -> LaTeXRender:
    return self._latex_render

  def use_asset(self, file_name: str, media_type: str) -> None:
    self._used_file_names[file_name] = media_type

  def add_asset(self, file_name: str, media_type: str, data: bytes) -> None:
    if file_name in self._used_file_names:
      return

    self._used_file_names[file_name] = media_type
    self._file.writestr(
      zinfo_or_arcname="OEBPS/assets/" + file_name,
      data=data,
    )

  @property
  def used_files(self) -> list[tuple[str, str]]:
    files: list[tuple[str, str]] = []
    for file_name in sorted(list(self._used_file_names.keys())):
      media_type = self._used_file_names[file_name]
      files.append((file_name, media_type))
    return files

  def add_used_asset_files(self) -> None:
    if self._assets_path is None:
      return
    for file_name in sorted(os.listdir(self._assets_path)):
      if file_name not in self._used_file_names:
        continue
      file_path = os.path.join(self._assets_path, file_name)
      self._file.write(
        filename=file_path,
        arcname="OEBPS/assets/" + file_name,
      )


from typing import Generator, Iterable
from enum import auto, Enum
from xml.etree.ElementTree import Element


ASSET_TAGS = ("figure", "table", "formula")

class AssetKind(Enum):
  FIGURE = auto()
  TABLE = auto()
  FORMULA = auto()

class AssetMatcher:
  def __init__(self):
    self._cloned_store: dict[AssetKind, list[Element]] = {}

  def register_raw_xml(self, root: Element) -> "AssetMatcher":
    for element in search_asset_tags(root):
      kind = self._tag_to_asset_kind(element.tag)
      cloned = self._clone_element(element)
      self._cloned_list(kind).append(cloned)
      element.clear()
    return self

  def register_virtual_dom(
        self,
        kind: AssetKind,
        hash: str | None = None,
        children: Iterable[Element] | None = None,
      ):
    tag_name = self._asset_kind_to_tag(kind)
    cloned = Element(tag_name)
    if hash is not None:
      cloned.attrib = { "hash": hash }
    if children is not None:
      for child in children:
        cloned.append(child)
    self._cloned_list(kind).append(cloned)

  def recover_asset_doms_for_xml(self, root: Element):
    for element in search_asset_tags(root):
      kind = self._tag_to_asset_kind(element.tag)
      cloned_list = self._cloned_store.get(kind, None)
      if cloned_list is None or len(cloned_list) == 0:
        continue
      cloned = cloned_list.pop(0)
      attrib = {
        **element.attrib,
        **cloned.attrib,
      }
      element.clear()
      element.attrib = attrib
      for child in cloned:
        element.append(child)

  def _tag_to_asset_kind(self, tag_name: str) -> AssetKind:
    if tag_name == "figure":
      return AssetKind.FIGURE
    elif tag_name == "table":
      return AssetKind.TABLE
    elif tag_name == "formula":
      return AssetKind.FORMULA
    else:
      raise ValueError(f"Unknown tag name: {tag_name}")

  def _asset_kind_to_tag(self, kind: AssetKind) -> str:
    if kind == AssetKind.FIGURE:
      return "figure"
    elif kind == AssetKind.TABLE:
      return "table"
    elif kind == AssetKind.FORMULA:
      return "formula"
    else:
      raise ValueError(f"Unknown asset kind: {kind}")

  def _clone_element(self, element: Element) -> Element:
    cloned = Element(element.tag, element.attrib)
    cloned.text = element.text
    cloned.tail = element.tail
    for child in element:
      cloned.append(self._clone_element(child))
    return cloned

  def _cloned_list(self, kind: AssetKind) -> list[Element]:
    cloned_list = self._cloned_store.get(kind, None)
    if cloned_list is None:
      cloned_list = []
      self._cloned_store[kind] = cloned_list
    return cloned_list

def search_asset_tags(target: Element) -> Generator[Element, None, None]:
  for child in target:
    if child.tag in ASSET_TAGS:
      yield child
    else:
      yield from search_asset_tags(child)


from io import TextIOWrapper
from typing import Callable
from dataclasses import dataclass
from resource_segmentation import Resource


@dataclass
class PageRef:
  page_index: int

@dataclass
class PageInfo:
  page_index: int
  main: Resource[PageRef]
  citation: Resource[PageRef] | None
  file: Callable[[], TextIOWrapper]


from typing import Literal


class I18N:
  def __init__(self, lan: Literal["zh", "en"]):
    if lan == "zh":
      self.unnamed: str = "未命名"
      self.cover: str = "封面"
      self.references: str = "引用"
    elif lan == "en":
      self.unnamed: str = "Unnamed"
      self.cover: str = "Cover"
      self.references: str = "References"


import os
import shutil

from json import dumps, loads
from typing import cast, Iterable, Callable
from xml.etree.ElementTree import tostring, fromstring, Element
from resource_segmentation import Resource, Incision

from ..llm import LLM
from ..pdf import PDFPageExtractor
from .common import PageInfo, PageRef
from .chunk_file import ChunkFile
from .ocr_extractor import extract_ocr_page_xmls
from .page import analyse_page
from .index import analyse_index, Index
from .citation import analyse_citations
from .main_text import analyse_main_texts
from .position import analyse_position
from .meta import extract_meta
from .chapter import generate_chapters
from .asset_matcher import ASSET_TAGS
from .window import parse_window_tokens, LLMWindowTokens
from .types import AnalysingStep, AnalysingProgressReport, AnalysingStepReport
from .utils import search_xml_and_indexes


def analyse(
  llm: LLM,
  pdf_page_extractor: PDFPageExtractor,
  pdf_path: str,
  analysing_dir_path: str,
  output_dir_path: str,
  window_tokens: LLMWindowTokens | int | None = None,
  report_step: AnalysingStepReport | None = None,
  report_progress: AnalysingProgressReport | None = None,
):
  state_machine = _StateMachine(
    llm=llm,
    pdf_page_extractor=pdf_page_extractor,
    pdf_path=pdf_path,
    analysing_dir_path=analysing_dir_path,
    output_dir_path=output_dir_path,
    window_tokens=parse_window_tokens(window_tokens),
    report_progress=report_progress,
    report_step=report_step,
  )
  state_machine.start()

_IndexXML = tuple[str, int, int]
_MARK_FILE_NAME = "MARK_DONE"

class _StateMachine:
  def __init__(
      self,
      llm: LLM,
      pdf_page_extractor: PDFPageExtractor,
      pdf_path: str,
      analysing_dir_path: str,
      output_dir_path: str,
      window_tokens: LLMWindowTokens,
      report_step: AnalysingStepReport | None,
      report_progress: AnalysingProgressReport | None,
    ):
    self._llm: LLM = llm
    self._pdf_page_extractor: PDFPageExtractor = pdf_page_extractor
    self._pdf_path: str = pdf_path
    self._analysing_dir_path: str = analysing_dir_path
    self._output_dir_path: str = output_dir_path
    self._window_tokens: LLMWindowTokens = window_tokens
    self._report_step: AnalysingStepReport | None = report_step
    self._report_progress: AnalysingProgressReport | None = report_progress
    self._index: Index | None = None
    self._index_did_load: bool = False
    self._pages: list[PageInfo] | None = None

  def start(self):
    self._run_analyse_step("ocr", self._extract_ocr)
    self._run_analyse_step("pages", self._analyse_pages)
    self._run_analyse_step("index", self._analyse_index)
    self._run_analyse_step("citations", self._analyse_citations)
    self._run_analyse_step("main_texts", self._analyse_main_texts)
    self._run_analyse_step("position", self._analyse_position)
    self._run_analyse_step("meta", self._extract_meta)
    self._generate_chapters()

  def _run_analyse_step(self, dir_name: str, func: Callable[[str], None]):
    mark_path = os.path.join(self._analysing_dir_path, dir_name, _MARK_FILE_NAME)
    if not os.path.exists(mark_path):
      dir_path = self._ensure_dir_path(
        dir_path=os.path.join(self._analysing_dir_path, dir_name),
      )
      func(dir_path)
      self._atomic_write(mark_path, "")

  def _extract_ocr(self, dir_path: str):
    assets_path = self._ensure_dir_path(os.path.join(self._analysing_dir_path, "assets"))
    index_xmls = self._list_index_xmls("page", dir_path)

    for page_index, page_xml in extract_ocr_page_xmls(
      extractor=self._pdf_page_extractor,
      pdf_path=self._pdf_path,
      expected_page_indexes=set(i for _, i, _ in index_xmls),
      cover_path=os.path.join(dir_path, "cover.png"),
      assets_dir_path=assets_path,
      report_step=self._report_step,
      report_progress=self._report_progress,
    ):
      self._atomic_write(
        file_path=os.path.join(dir_path, self._xml_name("page", page_index)),
        content=tostring(page_xml, encoding="unicode"),
      )

  def _analyse_pages(self, dir_path: str):
    from_path = os.path.join(self._analysing_dir_path, "ocr")
    done_page_indexes: set[int] = set()
    done_page_names: dict[int, str] = {}

    for file_name, page_index, _ in search_xml_and_indexes("page", dir_path):
      done_page_indexes.add(page_index)
      done_page_names[page_index] = file_name

    index_xmls = self._list_index_xmls("page", from_path)

    if self._report_step is not None:
      self._report_step(AnalysingStep.ANALYSE_PAGE, len(index_xmls))

    for i, (raw_name, page_index, _) in enumerate(index_xmls):
      if page_index not in done_page_indexes:
        raw_page_xml = self._read_xml(os.path.join(from_path, raw_name))
        previous_response_xml: Element | None = None
        if page_index > 0:
          file_name = done_page_names[page_index - 1]
          file_path = os.path.join(dir_path, file_name)
          previous_response_xml = self._read_xml(file_path)

        response_xml = analyse_page(
          llm=self._llm,
          raw_page_xml=raw_page_xml,
          previous_page_xml=previous_response_xml,
        )
        self._atomic_write(
          file_path=os.path.join(dir_path, raw_name),
          content=tostring(response_xml, encoding="unicode"),
        )
        done_page_names[page_index] = f"page_{page_index + 1}.xml"

      if self._report_progress is not None:
        self._report_progress(i + 1)

  def _analyse_index(self, dir_path: str):
    if self._report_step is not None:
      self._report_step(AnalysingStep.EXTRACT_INDEX, 0)

    from_path = os.path.join(self._analysing_dir_path, "pages")
    json_index, index = analyse_index(
      llm=self._llm,
      raw=(
        (i, self._read_xml(os.path.join(from_path, file_name)))
        for file_name, i, _ in self._list_index_xmls("page", from_path)
      )
    )
    if json_index is not None:
      self._atomic_write(
        file_path=os.path.join(dir_path, "index.json"),
        content=dumps(
          obj=json_index,
          ensure_ascii=False,
          indent=2,
        ),
      )
    if index is not None:
      self._index = index
      self._index_did_load = True

  def _analyse_citations(self, dir_path: str):
    with ChunkFile(dir_path) as file:
      analyse_citations(
        llm=self._llm,
        file=file,
        pages=self._load_pages(),
        data_max_tokens=cast(int, self._window_tokens.citations),
        tail_rate=0.15,
        report_step=self._report_step,
        report_progress=self._report_progress,
      )

  def _analyse_main_texts(self, dir_path: str):
    citations_dir_path = os.path.join(self._analysing_dir_path, "citations")
    with ChunkFile(dir_path) as file:
      analyse_main_texts(
        llm=self._llm,
        file=file,
        index=self._load_index(),
        pages=self._load_pages(),
        citations_dir_path=citations_dir_path,
        data_max_tokens=cast(int, self._window_tokens.main_texts),
        gap_rate=0.1,
        report_step=self._report_step,
        report_progress=self._report_progress,
      )

  def _analyse_position(self, dir_path: str):
    main_texts_path = os.path.join(self._analysing_dir_path, "main_texts")
    with ChunkFile(dir_path) as file:
      chunk_xmls = file.filter_origin_files(main_texts_path)
      if self._report_step is not None:
        chunk_xmls = list(chunk_xmls)
        self._report_step(AnalysingStep.MARK_POSITION, len(chunk_xmls))

      for i, (start_idx, end_idx, chunk_xml) in enumerate(chunk_xmls):
        position_xml = analyse_position(self._llm, self._load_index(), chunk_xml)
        file.atomic_write_chunk(start_idx, end_idx, position_xml)
        if self._report_progress is not None:
          self._report_progress(i + 1)

  def _extract_meta(self, dir_path: str):
    page_xmls: list[Element] = []
    page_dir_path = os.path.join(self._analysing_dir_path, "pages")
    head_count = 5

    if self._report_step is not None:
      self._report_step(AnalysingStep.ANALYSE_META, 0)

    for file_name, page_index, _ in search_xml_and_indexes("page", page_dir_path):
      if page_index >= head_count:
        break
      file_path = os.path.join(page_dir_path, file_name)
      page_xml = self._read_xml(file_path)
      page_xmls.append(page_xml)

    meta_json = extract_meta(self._llm, page_xmls)
    if meta_json is not None:
      meta_file_path = os.path.join(dir_path, "meta.json")
      self._atomic_write(
        file_path=meta_file_path,
        content=dumps(meta_json, ensure_ascii=False, indent=2),
      )

  def _generate_chapters(self):
    if self._report_step is not None:
      self._report_step(AnalysingStep.GENERATE_CHAPTERS, 0)

    if self._index is not None:
      file_path = os.path.join(self._output_dir_path, "index.json")
      self._atomic_write(
        file_path=file_path,
        content=dumps(self._index.json, ensure_ascii=False),
      )
    asset_hash_set: set[str] = set()

    for id, chapter_xml in generate_chapters(
      llm=self._llm,
      chunks_path=os.path.join(self._analysing_dir_path, "position"),
    ):
      if id is None:
        file_name = "chapter.xml"
      else:
        file_name = f"chapter_{id}.xml"

      self._atomic_write(
        file_path=os.path.join(self._output_dir_path, file_name),
        content=tostring(chapter_xml, encoding="unicode"),
      )
      content_xml = chapter_xml.find("content")
      for child in content_xml:
        if child.tag in ASSET_TAGS:
          hash = child.get("hash", None)
          if hash is not None:
            asset_hash_set.add(hash)

    self._copy_file(
      src_path=os.path.join(self._analysing_dir_path, "ocr", "cover.png"),
      dst_path=os.path.join(self._output_dir_path, "cover.png"),
    )
    self._copy_file(
      src_path=os.path.join(self._analysing_dir_path, "index", "index.json"),
      dst_path=os.path.join(self._output_dir_path, "index.json"),
    )
    self._copy_file(
      src_path=os.path.join(self._analysing_dir_path, "meta", "meta.json"),
      dst_path=os.path.join(self._output_dir_path, "meta.json"),
    )

    if len(asset_hash_set) > 0:
      asset_path = self._ensure_dir_path(os.path.join(self._output_dir_path, "assets"))
      for hash in asset_hash_set:
        src_path = os.path.join(self._analysing_dir_path, "assets", f"{hash}.png")
        dst_path = os.path.join(asset_path, f"{hash}.png")
        if os.path.exists(src_path):
          shutil.copy(src_path, dst_path)

  def _ensure_dir_path(self, dir_path: str) -> str:
    os.makedirs(dir_path, exist_ok=True)
    return dir_path

  def _load_index(self) -> Index | None:
    if not self._index_did_load:
      index_file_path = os.path.join(self._analysing_dir_path, "index", "index.json")
      self._index_did_load = True

      if os.path.exists(index_file_path):
        with open(index_file_path, "r", encoding="utf-8") as file:
          self._index = Index(loads(file.read()))

    return self._index

  def _list_index_xmls(self, kind: str, dir_path: str) -> list[_IndexXML]:
    index_xmls = list(search_xml_and_indexes(kind, dir_path))
    index_xmls.sort(key=lambda x: x[1])
    return index_xmls

  def _load_pages(self) -> list[PageInfo]:
    if self._pages is None:
      pages: list[PageInfo] = []
      pages_path = os.path.join(self._analysing_dir_path, "pages")

      for file_name, page_index, _ in self._list_index_xmls("page", pages_path):
        file_path = os.path.join(pages_path, file_name)
        page_xml = self._read_xml(file_path)
        page = self._parse_page_info(file_path, page_index, page_xml)
        pages.append(page)

      pages.sort(key=lambda p: p.page_index)
      self._pages = pages

    return self._pages

  def _parse_page_info(self, file_path: str, page_index: int, root: Element) -> PageInfo:
    main_children: list[Element] = []
    citation: Resource[PageRef] | None = None

    for child in root:
      if child.tag == "citation":
        citation = self._parse_text_info(page_index, child)
      else:
        main_children.append(child)

    return PageInfo(
      page_index=page_index,
      citation=citation,
      file=lambda: open(file_path, "rb"),
      main=self._parse_text_info(page_index, main_children),
    )

  def _parse_text_info(self, page_index: int, children: Iterable[Element]) -> Resource[PageRef]:
    # When no text is found on this page, it means it is full of tables or
    # it is a blank page. We cannot tell if there is a cut in the context.
    start_incision: Incision = Incision.UNCERTAIN
    end_incision: Incision = Incision.UNCERTAIN
    first: Element | None = None
    last: Element | None = None

    for child in children:
      if first is None:
        first = child
      last = child

    if first is not None and last is not None:
      if first.tag == "text":
        start_incision = self._attr_value_to_kind(first.attrib.get("start-incision"))
      if last.tag == "text":
        end_incision = self._attr_value_to_kind(last.attrib.get("end-incision"))

    tokens = self._count_elements_tokens(children)

    return Resource(
      count=tokens,
      start_incision=start_incision,
      end_incision=end_incision,
      payload=PageRef(page_index)
    )

  def _count_elements_tokens(self, elements: Iterable[Element]) -> int:
    root = Element("page")
    root.extend(elements)
    xml_content = tostring(root, encoding="unicode")
    return self._llm.count_tokens_count(xml_content)

  def _attr_value_to_kind(self, value: str | None) -> Incision:
    if value == "must-be":
      return Incision.MUST_BE
    elif value == "most-likely":
      return Incision.MOST_LIKELY
    elif value == "impossible":
      return Incision.IMPOSSIBLE
    elif value == "uncertain":
      return Incision.UNCERTAIN
    else:
      return Incision.UNCERTAIN

  def _xml_name(self, kind: str, page_index: int, page_index2: int | None = None) -> str:
    if page_index2 is None:
      return f"{kind}_{page_index + 1}.xml"
    else:
      return f"{kind}_{page_index + 1}_{page_index2 + 1}.xml"

  def _read_xml(self, file_path: str) -> Element:
    with open(file_path, "r", encoding="utf-8") as file:
      return fromstring(file.read())

  def _copy_file(self, src_path: str, dst_path: str):
    if os.path.exists(src_path):
      shutil.copy(src_path, dst_path)

  def _atomic_write(self, file_path: str, content: str):
    try:
      with open(file_path, "w", encoding="utf-8") as file:
        file.write(content)
        file.flush()
    except Exception as e:
      if os.path.exists(file_path):
        os.unlink(file_path)
      raise e



from __future__ import annotations
from typing import Generator
from dataclasses import dataclass
from xml.etree.ElementTree import Element
from ..llm import LLM
from .serial import serials, Citations
from .utils import search_xml_children


def generate_chapters(llm: LLM, chunks_path: str) -> Generator[tuple[int | None, Element], None, None]:
  session: _Session | None = None
  used_chapter_ids: set[int] = set()

  for serial in serials(llm, chunks_path):
    if session is None:
      session = _Session(None, serial.citations)
    else:
      session.update_serial_citations(serial.citations)

    for child in serial.main_texts:
      chapter_id = _try_to_take_chapter_id(child)
      if chapter_id is not None:
        if chapter_id in used_chapter_ids:
          continue
        if not session.is_empty:
          yield session.chapter_id, session.to_xml()
        session = _Session(chapter_id, serial.citations)
        used_chapter_ids.add(chapter_id)

      session.append(child)

  if session is not None and not session.is_empty:
    yield session.chapter_id, session.to_xml()

def _try_to_take_chapter_id(element: Element) -> int | None:
  if element.tag != "headline":
    return None
  id = element.attrib.pop("id", None)
  if id is None:
    return None
  return int(id)

@dataclass
class _Citation:
  label: str
  content: list[Element]

class _Session:
  def __init__(self, chapter_id: int | None, serial_citations: Citations ):
    self.chapter_id: int | None = chapter_id
    self._elements: list[Element] = []
    self._citations: list[tuple[int, _Citation]] = []
    self._refs: list[tuple[int, Element, Element]] = []
    self._next_ref_id: int = 1
    self._serial_citations: Citations = serial_citations

  @property
  def is_empty(self) -> bool:
    return len(self._elements) == 0

  def append(self, element: Element):
    for child, parent in search_xml_children(element):
      if child.tag == "ref":
        id = int(child.get("id"))
        self._refs.append((id, parent, child))
    self._elements.append(element)

  def update_serial_citations(self, citations: Citations):
    self._reset_ref_ids()
    self._refs.clear()
    self._serial_citations = citations

  def _reset_ref_ids(self):
    ids_map: dict[int, int] = {}
    for origin_id, parent, ref in self._refs:
      citation = self._serial_citations.get(origin_id)
      if citation is None:
        self._remove_ref(parent, ref)
      else:
        new_id = ids_map.get(origin_id, None)
        if new_id is None:
          new_id = self._next_ref_id
          self._next_ref_id += 1
          ids_map[origin_id] = new_id
          self._citations.append((new_id, _Citation(
            label=citation.label,
            content=citation.content,
          )))
        ref.set("id", str(new_id))

  def _remove_ref(self, parent: Element, ref: Element):
    pre_element: Element | None = None
    for child in parent:
      if child == ref:
        break
      pre_element = child
    if ref.tail is not None:
      if pre_element is not None:
        pre_element.tail = (pre_element.tail or "") + ref.tail
      else:
        parent.text = (parent.text or "") + ref.tail
    parent.remove(ref)

  def to_xml(self) -> Element:
    self._reset_ref_ids()
    chapter_xml = Element("chapter")
    content_xml = Element("content")
    chapter_xml.append(content_xml)
    for element in self._elements:
      content_xml.append(element)

    if len(self._citations) > 0:
      citations_xml = Element("citations")
      chapter_xml.append(citations_xml)
      for id, citation in self._citations:
        citation_xml = Element("citation")
        citation_xml.set("id", str(id))
        citations_xml.append(citation_xml)
        label_xml = Element("label")
        label_xml.text = citation.label
        citation_xml.append(label_xml)
        for child in citation.content:
          citation_xml.append(child)

    return chapter_xml
